/***************************************************************************
                          leveldaten.cpp  -  description
                             -------------------
    begin                : Fri Jul 21 2000
    copyright            : (C) 2000 by Immi
    email                : cuyo@pcpool.mathematik.uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <cstdlib>

#include <qbitmap.h>
#include <qpainter.h>

#include "cuyointl.h"
#include "fehler.h"
#include "pfaditerator.h"
#include "leveldaten.h"
#include "datendatei.h"
#include "knoten.h"

#include "code.h"
#include "blop.h"



#define toptime_default 50


/** Wird in main.cpp definiert. Ist gepfuscht. */
extern __String gLevelDatei;



/* Globale Variable mit den Level-Daten */

LevelDaten * ld;

/** */
LevelDaten::LevelDaten(): mLevelConf(new DatenDatei()),
  mGeladen(false),
/* Wir basteln uns ein Array, dessen Indizierung nicht mit 0 beginnt...: */
  mSorten(mSortenIntern - blopart_min_sorte)
/* Uninitialisierte Blops entstehen sehr frh, und die haben auch schon
   einen Bildstapel. Und der greift auf mStapelHoehe() zu. */
 // mStapelHoehe(0)
{
  ld = this;
  mIntLevelNamen.setAutoDelete(true);
  ladLevelConf();
}


/** */
LevelDaten::~LevelDaten() {
  delete mLevelConf;
}



void LevelDaten::ladLevelConf() {

  /* Jetzt ist erst mal nix mehr geladen. Erst wenn ladLevelConf()
     durchgelaufen ist, ist wieder was da. */
  mGeladen = false;
  
  /* Falls schon mal was geladen wurde: Erst mal alles
     wieder rausschmeien. */
  mLevelConf->leeren();

  /* Level-descr-Datei ffnen */
  
  /* in verschiedenen Pfaden suchen...
     (Der Pfaditerator throwt ggf.)
     true bedeutet: Default-Pfad auf Ort der level descr setzen.
     LEVELDESCR wird (im Moment) in inkompatibel.h definiert. */
  PfadIterator pi(LEVELDESCR, true);
  for (; !QFile(pi.pfad()).exists(); ++pi);

  mLevelConfPfad = pi.pfad();

  /* Whrend laden() wird mIntLevelNamen ausgefllt. Falls da vorher
     noch Mll drin war, lschen wir das. (Ist hier eigentlich
     unntig; sieht aber sauberer aus. */
  mIntLevelNamen.clear();

  /* Hier findet das parsen statt. */
  mLevelConf->laden(mLevelConfPfad);



  /* Hat der Benutzer noch eine eigene Leveldatei angegeben?
     Dann wird die jetzt auch noch geladen. */
  if (gLevelDatei != "") {
    
    if (!QFile(gLevelDatei).exists())
      throw Fehler(_("Could not load \"%s\"."), gLevelDatei.data());
    
    /* Wenn der Benutzer eine Datei angegeben hat, sollen nur die Level
       aus seiner Datei Level werden. Also die bisherige Level-Liste
       wieder lschen. */
     mIntLevelNamen.clear();
    
    /* Hier findet schon wieder parsen statt. */
    mLevelConf->laden(gLevelDatei);

  } else {
  
    /* Benutzer hat keine eigene Datei angegeben. Dann holen wir
       uns doch die Level-Liste aus "level=..." */
    mLevelConf->setEinSpielerVersion(false);
    verlangeSchluessel("level");
    ListenKnoten * lena = mLevelConf->getListenEintrag("level");
    int l = lena->getLaenge();
    mIntLevelNamen.resize(l);
    for (int i = 0; i < l; i++)
      mIntLevelNamen.insert(i, new __String(lena->getWort(i)));
  }

  /* OK, Laden war erfolgreich. */
  mGeladen = true;
}




/** Wird whrend des Parsens (d. h. innerhalb von ladLevelConf() von
    DefKnoten aufgerufen, wenn ein neuer Level gefunden wurde. Daraus
    kann eine Level-Liste gebastelt werden, wenn "level=..." nicht
    verwendet werden soll. */
void LevelDaten::levelGefunden(__String lna) {
  int l = mIntLevelNamen.size();
  mIntLevelNamen.resize(l + 1);
  mIntLevelNamen.insert(l, new __String(lna));
}




/** fllt alle Daten in diesem Objekt fr Level nr aus; throwt bei Fehler */
void LevelDaten::ladLevel(int nr, int spz) {

  if (!mGeladen)
    throw Fehler(_("Sorry, no working level description file available."));

  /* In den obersten Abschnitt der level descr springen. Wir knnten uns
     woanders befinden, wenn es irgend wann mal einen Fehler gegeben hatte. */
  mLevelConf->initSquirrel();

  /* Nur fr den Fall eines frhen throws... */
  mLevelName = "";

  try {
  
    mIntLevelName = getIntLevelName(nr);

    /* In den Abschnitt dieses Levels springen. (Springt automatisch
       bei } wieder raus. */
    DatenDateiPush ddp(*mLevelConf, mIntLevelName);
    
    /* Spielerzahlabhngige Unterschiede bercksichtigen */
    mSpielerZahl = spz;
    mLevelConf->setEinSpielerVersion(spz == 1);

    /* DefKnoten des Levels abspeichern. */
    mLevelKnoten = mLevelConf->getSquirrelPos();

    /* Level-Name */
    verlangeSchluessel("name");
    mLevelName = mLevelConf->getEintrag("name");

    /* Level-Autor */
    verlangeSchluessel("author");
    mLevelAutor = mLevelConf->getEintrag("author");

    /* Beschreibungstext (optional) */
    mBeschreibung = mLevelConf->getEintrag("description", "");

    /* Wie viele Steine mssen zusammen, damit sie platzen? */
    verlangeSchluessel("numexplode");
    mPlatzAnzahl = mLevelConf->getZahlEintrag("numexplode");
    if (mPlatzAnzahl < 1)
      throw Fehler(_("numexplode < 1"));

    /* Hintergrundfarbe... (optional; Default: wei)
       Achtung: Die Hintergrundfarbe muss gesetzt werden, _bevor_
       Bildchen geladen werden, da es als Bonus-Farbe im XPM
       "Background" gibt... (im Moment nur fr Explosion sinnvoll) */
    mHintergrundFarbe = mLevelConf->getFarbEintrag("bgcolor", __white);

    /* Hintergrundbilchen (optional) */
    mMitHintergrundbildchen = mLevelConf->hatEintrag("bgpic");
    if (mMitHintergrundbildchen)
      mHintergrundBild.laden(mLevelConf->getEintrag("bgpic"));

    /* Schriftfarbe... (optional; Default: schwarz) */
    schriftFarbe = mLevelConf->getFarbEintrag("textcolor", __black);

    /* Hetzrandfarbe... (optional; Default: hellgrau) */
    hetzrandFarbe = mLevelConf->getFarbEintrag("topcolor", __lightGray);

    /* Hetzrandgeschwindigkeit (optional) */
    hetzrandZeit = mLevelConf->getZahlEintrag("toptime", toptime_default);
    if (hetzrandZeit < 1)
      throw Fehler(_("toptime < 1"));

    /* Hetzrandbildchen (optional) */
    mMitHetzbildchen = mLevelConf->hatEintrag("toppic");
    if (mMitHetzbildchen) {
      mHetzBild.laden(mLevelConf->getEintrag("toppic"));
  	
      /* Hetzrandberlapp (optional) */
      mHetzrandUeberlapp =
	mLevelConf->getZahlEintrag("topoverlap", mHetzBild.getHoehe());
    } else
      mHetzrandUeberlapp = 0;

    /* Gras nur bei Kettenreaktion? (optional) */
    mGrasBeiKettenreaktion = mLevelConf->getZahlEintrag("chaingrass", 0);

    /* Senkrecht spiegeln? (optional) */
    mSpiegeln = mLevelConf->getZahlEintrag("mirror");

    /* Andere Nachbarschaft? (optional) */
    mNachbarschaft =
      mLevelConf->getZahlEintrag("neighbours", nachbarschaft_normal);
    if (mNachbarschaft < 0 || mNachbarschaft > nachbarschaft_letzte)
      throw Fehler(_("neighbours out of range"));
    /* Sechseck-Raster? */
    mSechseck = mNachbarschaft == nachbarschaft_6 ||
      mNachbarschaft == nachbarschaft_6_schraeg;


    /* Zufllige Graue? (optional) */
    mZufallsGraue =
      mLevelConf->getZahlEintrag("randomgreys", zufallsgraue_keine);
  	  	
    /* Wo sind welche Grasbildchen am Anfang? */
    verlangeSchluessel("startdist");
    mAnfangsZeilen = mLevelConf->getListenEintrag("startdist");

    /* Wie viele Farben gibts in diesem Level? Und wie heien die Bildchen? */
    ListenKnoten * picsnamen;
    verlangeSchluessel("pics");
    picsnamen = mLevelConf->getListenEintrag("pics");
    mAnzFarben = picsnamen->getLaenge();
    if (mAnzFarben < 1)
      throw Fehler(_("#pics < 1"));
  	
    if (mAnzFarben > max_farben_zahl) {
      __String s;
      s.sprintf(_("#pics > %d"), max_farben_zahl);
      throw Fehler(s);
    }


    /***** Noch ein paar einzelne Bilder laden *****/
    
    /* Explosion laden. */
    mExplosionBild.laden("explosion.xpm");

    /* Ziffern fr Debug-Ausgabe laden. */
    mZiffernBild.laden("ziffern.xpm");

    

    /***** Blops laden *****/

    /* Wie viele Bilder malt ein Blop in einem Schritt hchstens?
       Erst mal keine. Die Sorten erhhen diese Variable selbst, wenn
       man sie ldt. */
    mStapelHoehe = 0;

    /* Farb-Bilder laden */
    for (int bnr = 0; bnr < mAnzFarben; bnr++)
      mSorten[bnr].laden(picsnamen->getWort(bnr), blopart_farbe);

    /* Grau-Bildchen laden */
    /* Achtung: "greypic" steht auch in knoten.cpp */
    verlangeSchluessel("greypic");
    mSorten[blopart_grau].laden(mLevelConf->getEintrag("greypic"),
                                blopart_grau);

    /* Grasbildchen laden */
    /* Achtung: "startpic" steht auch in knoten.cpp */
    verlangeSchluessel("startpic");
    mSorten[blopart_gras].laden(mLevelConf->getEintrag("startpic"),
                                blopart_gras);

    /* Leer-Bildchen (optional) */
    mMitLeerBildchen = mLevelConf->hatEintrag("emptypic");
    /* Auch, wenn es kein Leer-Bildchen gibt, soll es geladen werden;
       dann liefert getEintrag() "" zurck, und Sorte::laden() wei,
       dass es nur alle Werte auf Defaults setzen soll.
       Letztes true: Per default mit Rand verbinden */
    mSorten[blopart_keins].laden(mLevelConf->getEintrag("emptypic"),
                                 blopart_keins);

    /* Globaler Code (optional) */
    mSorten[blopart_global].laden("global", blopart_global);
    
    
    /* Ok, alle Sorten geladen. Jetzt stimmt die Variable
       mStapelHoehe. Ab jetzt drfen also Blops erzeugt werden
       (wenn man mchte). */

   //fprintf(stderr, "mStapelHoehe = %d\n", mStapelHoehe);    


  } catch (Fehler f) {
    __String fs;
    fs.sprintf(_("Error in \"%s\""),
	       mLevelConfPfad.data());
    if (!mLevelConf->getSquirrelPosString().isEmpty())
      fs += ", Section " + mLevelConf->getSquirrelPosString();
    fs += ":\n" + f.mText + "\n";
    if (!mLevelName.isEmpty())
      fs += _("(Level \"") + mLevelName + _("\")\n");
    throw Fehler(fs);
  }
	
}





/** liefert zurck, wie viel Gras am Anfang des Levels
    da ist */
int LevelDaten::getGrasAnzahl() {
  int anz_grasfelder = 0;
  /* Zeilen durchgehen... */
  for (int i = 0; i < mAnfangsZeilen->getLaenge(); i++) {
    __String zeile = mAnfangsZeilen->getWort(i);
    for (int x = 0; x < (int) zeile.length(); x++) {
      /* Etwas provisoritsch... */
      if (zeile.data()[x] >= 'A' && zeile.data()[x] <= 'Z')
	anz_grasfelder++;
    }
  }
  return anz_grasfelder;
}


/** Throwt, wenn der angegebene Schlssel in level descr nicht existiert. */
void LevelDaten::verlangeSchluessel(const char * s) const {
  if (!mLevelConf->hatEintrag(s)) {
    __String t;
    t.sprintf(_("Key \"%s\" is missing"), s);
    throw Fehler(t);
  }
}



/** Sollte am Anfang des Levels aufgerufen werden; kmmert sich
    um den Global-Blop */
void LevelDaten::startLevel() const {
  Blop::gGlobalBlop = Blop(blopart_global);
}


/** Sollte einmal pro Spielschritt aufgerufen werden (bevor
    Spielfeld::spielSchritt() aufgerufen wird). Kmmert sich 
    um den Global-Blop */
void LevelDaten::spielSchritt() const {
  Blop::beginGleichzeitig();
  Blop::gGlobalBlop.animiere();
  Blop::endGleichzeitig();
}


/** Liefert zurck, wie viele Level es gibt. */
int LevelDaten::getLevelAnz() const {
  if (!mLevelConf)
    throw Fehler(_("Sorry, no working level description file available."));
  return mIntLevelNamen.size();
}

/** Liefert den internen Namen von Level nr zurck. */
__String LevelDaten::getIntLevelName(int nr) const {
  return nr == level_titel ?
    __String("Title")
    :
    *mIntLevelNamen.at(nr - 1);
}

/** Liefert den Namen von Level nr zurck. Liefert "???" bei Fehler. */
__String LevelDaten::getLevelName(int nr) const {
  try {
    DatenDateiPush ddp(*mLevelConf, getIntLevelName(nr));

    verlangeSchluessel("name");
    return mLevelConf->getEintrag("name");

  } catch (Fehler f) {
    return "???";
  }
}

